#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import ctypes
import logging, sys

''' insure ctypes basic types are subverted '''
from haystack import model

from haystack.model import is_valid_address,is_valid_address_value,getaddress,array2bytes,bytes2array
from haystack.model import LoadableMembers,RangeValue,NotNull,CString, IgnoreMember

import ctypes_nss_generated as gen

log=logging.getLogger('ctypes_nss')


# ============== Internal type defs ==============

'''
http://mxr.mozilla.org/firefox/source/security/nss/lib/ssl/ssl.h
http://mxr.mozilla.org/firefox/source/security/nss/lib/ssl/sslt.h
http://mxr.mozilla.org/firefox/source/security/nss/lib/ssl/sslimpl.h#912
'''

class NSSStruct(LoadableMembers):
  ''' defines classRef '''
  pass



# replace c_char_p with our String handler
if type(gen.STRING) != type(CString):
  print 'STRING is not model.CString. Please correct ctypes_nss_geenrated with :' 
  print 'from model import CString' 
  print 'STRING = CString' 
  import sys
  sys.exit()


################ START copy generated classes ##########################

# copy generated classes (gen.*) to this module as wrapper
model.copyGeneratedClasses(gen, sys.modules[__name__])

# register all classes (gen.*, locally defines, and local duplicates) to haystack
# create plain old python object from ctypes.Structure's, to picke them
model.registerModule(sys.modules[__name__])

################ END   copy generated classes ##########################




############# Start expectedValues and methods overrides #################



sslSocket.expectedValues = {
 "fd": RangeValue(1,0xffff), 
  #"version": [0x0002, 0x0300, 0x0301 ], #sslproto.h
  "version": RangeValue(1,0x0301), 
  "clientAuthRequested": [0,1], # 
  "delayDisabled": [0,1], # 
  "firstHsDone": [0,1], # 
  "handshakeBegun": [0,1], # 
  "TCPconnected": [0,1], # 
  "lastWriteBlocked": [0,1], # 
  "url": NotNull,
  }



sslOptions.expectedValues = {
#  "useSecurity": [0,1], # doh... 
  "handshakeAsClient" : [1],
  "handshakeAsServer" : [0],
}

sslSecurityInfo.expectedValues = {
  "cipherType": RangeValue(1,0xffffffff),
  #"writeBuf" : IgnoreMember,
}


ssl3State.expectedValues = {
}


ssl3CipherSpec.expectedValues = {
  "cipher_def" : NotNull,
  "mac_def" : NotNull,
}


'''
# set expected values 
SSLCipherSuiteInfo.expectedValues={
  "cipherSuite": RangeValue(0,0x0100), # sslproto.h , ECC is 0xc00
  "authAlgorithm": RangeValue(0,4),
  "keaType": RangeValue(0,4),
  "symCipher": RangeValue(0,9),
  "macAlgorithm": RangeValue(0,4),
}

SSLChannelInfo.expectedValues={
  'compressionMethod':RangeValue(0,1),
}

SECItem.expectedValues={
  'type':RangeValue(0,15), # security/nss/lib/util/seccomon.h:64
}

SECKEYPublicKey.expectedValues={
  'keyType':RangeValue(0,6), # security/nss/lib/cryptohi/keythi.h:189
}

#NSSLOWKEYPublicKey.expectedValues={
#  'keyType':RangeValue(0,5), # security/nss/lib/softoken/legacydb/lowkeyti.h:123
#}

CERTCertificate.expectedValues={ # XXX TODO CHECK VALUES security/nss/lib/certdb/certt.h
  'keyIDGenerated': [0,1], # ok
  'keyUsage': [RangeValue(0,gen.KU_ALL), # security/nss/lib/certdb/certt.h:570
      gen.KU_KEY_AGREEMENT_OR_ENCIPHERMENT ,
      RangeValue(gen.KU_NS_GOVT_APPROVED, gen.KU_ALL | gen.KU_NS_GOVT_APPROVED),
      gen.KU_KEY_AGREEMENT_OR_ENCIPHERMENT | gen.KU_NS_GOVT_APPROVED ] ,
  'rawKeyUsage': [RangeValue(0,gen.KU_ALL), # security/nss/lib/certdb/certt.h:570
      gen.KU_KEY_AGREEMENT_OR_ENCIPHERMENT ,
      RangeValue(gen.KU_NS_GOVT_APPROVED, gen.KU_ALL | gen.KU_NS_GOVT_APPROVED),
      gen.KU_KEY_AGREEMENT_OR_ENCIPHERMENT | gen.KU_NS_GOVT_APPROVED ] ,
  'keyUsagePresent': [0,1], # ok
  'nsCertType': [ RangeValue(0,0xff), # security/nss/lib/certdb/certt.h:459
      RangeValue(gen.EXT_KEY_USAGE_TIME_STAMP , gen.EXT_KEY_USAGE_TIME_STAMP | 0xff ), # not sure
      RangeValue(gen.EXT_KEY_USAGE_STATUS_RESPONDER  , gen.EXT_KEY_USAGE_STATUS_RESPONDER | 0xff ), # not sure
      RangeValue(gen.EXT_KEY_USAGE_TIME_STAMP | gen.EXT_KEY_USAGE_STATUS_RESPONDER  , 
                  gen.EXT_KEY_USAGE_TIME_STAMP | gen.EXT_KEY_USAGE_STATUS_RESPONDER | 0xff ) ] , # not sure
  'keepSession': [0,1], # ok
  'timeOK': [0,1], # ok
  'isperm': [0,1], # ok
  'istemp': [0,1], # ok
  'isRoot': [0,1], # ok
  'ownSlot': [0,1], # ok
  'referenceCount': RangeValue(1,0xffff), # no idea... should be positive for sure. more than 65535 ref ? naah
  'subjectName': NotNull, # a certificate must have one...
}

CERTSignedCrl.expectedValues = {
  'isperm': [0,1], # ok
  'istemp': [0,1], # ok
  'referenceCount': RangeValue(1,0xffff), # no idea... should be positive for sure. more than 65535 ref ? naah
}

CERTCertTrustStr.expectedValues = {
  #'sslFlags': [1<<i for i in range(0,11+1)] , # ok security/nss/lib/softoken/legacydb/pcertt.h:438
  'sslFlags': RangeValue(0, 1<<12) , # simpler
  'emailFlags': RangeValue(0, 1<<12) , # simpler
  'objectSigningFlags': RangeValue(0, 1<<12) , # simpler
}

'''

##########


def printSizeof(mini=-1):
  for (name,klass) in inspect.getmembers(sys.modules[__name__], inspect.isclass):
    if type(klass) == type(ctypes.Structure) and klass.__module__ == 'ctypes_nss_generated' :
      if ctypes.sizeof(klass) > mini:
        print '%s:'%name,ctypes.sizeof(klass)
  #print 'SSLCipherSuiteInfo:',ctypes.sizeof(SSLCipherSuiteInfo)
  #print 'SSLChannelInfo:',ctypes.sizeof(SSLChannelInfo)

if __name__ == '__main__':
  printSizeof()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import logging, os, socket, sys, time
import multiprocessing, Queue

from lrucache import LRUCache

WAIT_RETRANSMIT = 20
QSIZE = 5000

log=logging.getLogger('stream')


def hexify(data):
  s=''
  for i in range(0,len(data)):
    s+="%02x"% ord(data[i])
    if i%16==15:
      s+="\r\n"
    elif i%2==1:
      s+=" "
  #s+="\r\n"
  return s


class MissingDataException (Exception):
  ''' declare we are lost in translation '''
  def __init__(self,nb=0):
    self.nb=nb

class BadStateError (Exception):
  pass


class State:
  pass


class TCPState(State):
  ''' TCP state. 
    enqueueRaw is used by TCP Stream to triage packet in one direction.
    STATE SEARCH: packets are then ordered into orderedQueue following a simple TCP seq state machine
    STATE ACTIVE: packets payload are added to the socket in an ordered fashion following a simple TCP seq state machine
    orderedQueue contains ordered packets waiting to be processed
      - at some point, packets data from ordered Queue must be put in write_socket
      - a override can change condition of write.... (decryption testing ?) 
      - setSearchMode() set the state to queue packest
      - setActiveMode() set the state to write packets payload
    write_socket is used to put data from ordered TCP packets.
    read_socket is used by a reader ( Packetizer) to read data in active mode.

  '''
  name = None
  packet_count = 0
  byte_count = 0
  start_seq = None
  max_seq = 0
  expected_seq = 0
  rawQueue = None
  orderedQueue = None
  write_socket = None
  read_socket = None
  ts_missing = None
  # for variable size retransmission
  packets = None
  activeLock = None
  def __init__(self,name):
    self.name=name
    self.rawQueue = {}
    self.orderedQueue = Queue.Queue(QSIZE)
    self.packets = LRUCache(QSIZE)
    self.activeLock   = multiprocessing.Lock()
    # make socket UNIX way. non existent in Windows
    read, write = socket.socketpair()
    self.read_socket  = socket.socket(_sock=read) ## useless to get a python object we do not override after all?
    self.write_socket = write
    # ok
    self.setSearchMode()
    log.debug('%s: created in search mode'%(self.name))
    return 

  def _enqueueRaw(self, packet):
    ''' the rawQueue gets all unexpected packets. 
    reodering will happen before adding them to orderedQueue or socket '''
    self.rawQueue[packet.seq]= packet
    return

  def _isMissing(self):
    return not (self.ts_missing is None)
  def _resetMissing(self):
    self.ts_missing = None
  def _setMissing(self):
    self.ts_missing = time.time()


  def _requeue(self):
    ''' Internal func .
        get packets from rawQueue and put them in processing orderedQueue or in socket '''
    queue = self.rawQueue.values()
    queue.sort(key=lambda p: p.seq)
    # add the first and the next that are in perfect suite.
    toadd = [queue.pop(0),]
    log.debug('%s: _requeue : Poped packet seq : %s'%(self.name, toadd[0].seq))
    for i in xrange(0,len(queue)):
      if toadd[-1].seq + len(toadd[-1].payload) == queue[0].seq :
        toadd.append(queue.pop(0))
      else:
        log.debug('Stop prequeuing  toadd[-1].seq+len(toadd[-1].payload): %d , queue[0].seq: %d'%(toadd[-1].seq+len(toadd[-1].payload) , queue[0].seq ))
        break
    # let remaining in state.queue
    self.rawQueue=dict( [ (p.seq, p) for p in queue])
    # add to output 
    for p in toadd:
      self.addPacket( p )
    log.debug('Prequeued %d packets, remaining %d, queued from %d to %d '%(
                        len(toadd), len(self.rawQueue), toadd[0].seq, (toadd[-1].seq + len(toadd[-1].payload)) 
                        ))
    # set expected 
    self.max_seq = toadd[-1].seq
    self.expected_seq = toadd[-1].seq + len(toadd[-1].payload) 
    # reset time counter
    if len(self.rawQueue) > 0:
      self._setMissing()
    else:
      self._resetMissing()
    return

  def _addPacketToOrderedQueue(self, packet ):
    ''' search mode '''
    ''' Ordered packet are added here before their contents gets to socket. '''
    log.debug('%s: _addPacketToOrderedQueue'%(self.name))
    self.orderedQueue.put(packet)
    return 1

  def _addPacketToSocket(self, packet):
    ''' active mode '''
    self.activeLock.acquire()
    cnt = self.write_socket.send( packet.payload.load )
    self.byte_count   += cnt
    self.packet_count += 1
    self.activeLock.release()
    log.debug('%s: Adding packet to socket - %d bytes added'%(self.name, cnt))
    return True

  def _checkStateFalse(self, packet):
    return False

  def _checkState1(self, packet):
    ''' temp methods to get initial seqnum from first packet '''
    seq = packet.seq    
    ## debug head initialisation
    if self.start_seq is None:
      self.start_seq=seq
      self.expected_seq=seq
      # head done. switch to normal behaviour
      self.checkState = self._checkState
      log.debug('%s: Switching to regular checkState'%(self.name))
    return self._checkState( packet )
  
  def _checkState(self, packet):
    ''' if packet is expected return True.
        else, add it to queues and return False
    '''
    ## we should not processs empty packets
    payloadLen = len(packet.payload)
    if payloadLen == 0:
      return False
    seq = packet.seq
    log.debug('%s: Checking state of packet %d exp: %d len: %d'%(self.name, packet.seq, self.expected_seq, len(packet.payload.load)))
    
    # packet is expected 
    if seq == self.expected_seq: # JIT
      log.debug('%s: got a good paket, adding..'%(self.name))
      self.max_seq = seq
      self.expected_seq = seq + payloadLen
      self.addPacket(packet)
      # 
      self.packets[seq] = packet # debug the packets
      self._resetMissing()
      # check if next is already in self.queue , if expected has changed
      self.checkForExpectedPackets()
      return True

    # packet is future
    elif seq > self.expected_seq:
      log.debug('%s: Future packet, queuing it...'%(self.name)) 
      # seq is in advance, add it to queue
      self._enqueueRaw(packet)
      #log.debug('Queuing packet seq: %d len: %d %s'%(seq, payloadLen, state))

      # we are waiting for something...
      if not self._isMissing():
        self._setMissing()

      # check if next is already in self.queue 
      self.checkForExpectedPackets()
      #do not add this one, it's maybe already added anyway...
      return False

    # packet is a retranmission 
    elif seq < self.expected_seq:
      # TCP retransmission
      log.debug('TCP retransmit - We just received %d when we already processed %d'%(seq, self.max_seq))
      # never hitted
      if seq+payloadLen > self.expected_seq :
        log.warning(' ***** EXTRA DATA FOUND ON TCP RETRANSMISSION ')
        # we need to recover the extra data to put it in the stream
        nb   = (seq+payloadLen) - self.expected_seq
        data = packet.payload.load[-nb:]
        log.warning('packet seq %d has already been received'%(seq))
        log.warning('recent  packet : %s'%(repr(packet.underlayer) ))
        log.warning('first   packet : %s'%(repr(self.packets[seq].underlayer ) ))
        seq2 = seq+len(self.packets[seq].payload)
        log.warning('first+1 packet : %s'%(repr(self.packets[seq2].underlayer ) ))
        packet.payload.load = data
        log.warning('NEW     packet : %s'%(repr(packet.underlayer) ))
        # updates seq
        self.max_seq = seq
        self.expected_seq = seq2
        # save it
        self.packets[seq]=packet 
        # we can process this (new) one, it's containing only the non-used remains
        self.addPacket(packet)        
        return True
      # ignore it, it's a retransmission
      return False
    # else, seq < expected_seq and seq >= self.max_seq. That's not possible... it's current packet.
    # it's a dup already dedupped ? partial fragment ?
    log.warning('received a bogus fragment seq: %d for %s'%(seq, self))
    return False
    
  ################ PUBLIC METHODS 
  ''' check the state of the packet.
  if packet is expected return True.
        else, add it to queues and return False 
  at first, we are expected to find our first seq '''
  checkState = _checkState1
  ''' at first, add to queue. When the state goes into active mode, switch to socket '''
  addPacket = _addPacketToOrderedQueue

  def checkForExpectedPackets(self):
    ''' check for some internal expectation. '''
    ret = False
    log.debug('time to check the raw queue for expected packets ')
    if self.expected_seq in self.rawQueue : 
      log.debug('requeue all expected packets to ordered queue ')
      self._requeue()
      ret = True
    # waiting for too long
    elif self._isMissing() and  time.time() > ( self.ts_missing + WAIT_RETRANSMIT) : 
      log.error('%s: Some data is missing. the sniffer losts some packets ? Dying. '%(self.name))      
      #raise MissingDataException()
      self.checkState = self._checkStateFalse
    return ret

  def getSocket(self):
    return self.read_socket

  def getFirstPacketData(self, block=False):
    ''' pop the first packet '''
    if not self.searchMode:
      raise BadStateError('State %s is in Active Mode. Not poping allowed')
    # wait for it     if self.orderedQueue.qsize() == 0: or except Queue.Empty
    p = self.orderedQueue.get(block=block)
    d = p.payload.load
    self.orderedQueue.task_done()
    #log.info('orderedQueue size : %d'%(self.orderedQueue.qsize()))
    return d, self.orderedQueue.qsize()
  
  def setSearchMode(self):
    self.addPacket = self._addPacketToOrderedQueue
    self.searchMode = True

  def setActiveMode(self, data=None):
    ''' go in active mode.
      all ordered packets will be written to the socket for subsequent use...
      
      WARNING: you should have a thread running on read_socket, otherwise, 
      the socket buffer space will quickly be overflown and this will block on socket.send()
      
    @param data:  some data can be pre-written to the socket  .
    '''
    # stop any data from behing inserted in socket witouht proper timing
    log.debug('Activating the active mode. Data will be decrypted')
    self.activeLock.acquire()
    self.searchMode = False
    self.addPacket = self._addPacketToSocket
    if data is not None:
      log.debug('Prepended %d bytes of data before remaining packets'%(len(data) ) )
      self.byte_count   += self.write_socket.send( data )
      self.packet_count += 1
    # push data
    queue = self.orderedQueue
    self.orderedQueue = Queue.Queue(QSIZE)
    while not queue.empty():
      packet = queue.get()
      self.byte_count   += self.write_socket.send( packet.payload.load )
      self.packet_count += 1
    log.debug('%d bytes written for %d packets'%(self.byte_count, self.packet_count))
    self.activeLock.release()
    # operation can now resume. the socket is active
    return 
  
  def finish(self):
    ''' use current data to the best of it's knowledge .. '''
    self.orderedQueue.join()
    # if orderedQueue is empty, data is in socket.
    # no more data in coming up from stream/network we can close the socket.
    self.activeLock.acquire()
    self.write_socket.close()
    self.activeLock.release()
    
  def __str__(self):
    return "%s: %d bytes/%d packets max_seq:%d expected_seq:%d q:%d"%(self.name, self.byte_count,self.packet_count,
                self.max_seq,self.expected_seq, self.orderedQueue.qsize())
  
class stack:
  ''' A stream is duplex. '''
  def __init__(self):
    self.inbound=TCPState('inbound')
    self.outbound=TCPState('outbound')
  def __str__(self):
    return "\n%s\n%s"%(self.inbound,self.outbound)
  
class Stream:
  ''' 
    A Stream is a duplex communication.
    
    All packets coming from a scapy socket are triaged here, for both directions.
    
    If the TCPState is in active mode (setActiveMode()) data will be available in the socket.
    
  '''
  worker=None
  def __init__(self, inQueue, connection, protocolName):
    ''' 
    @param inQueue: packet Queue from socket_scapy   ## from multiprocessing import Process, Queue
    @param connectionTuple: connection Metadata to identify inbound/outbound
    '''
    self.inQueue = inQueue # triage must happen 
    self.connection = connection
    self.protocolName = protocolName
    # contains TCP state & packets queue before reordering    
    self.stack = stack()  # duplex context
    self.running = True

  def getInbound(self):
    return self.stack.inbound
  
  def getOutbound(self):
    return self.stack.outbound

  def _isInbound(self, packet):
    raise NotImplementedError()

  def check(self):
    return self.running

  def pleaseStop(self):
    self.running = False
   
  def run(self):
    ''' loops on self.inQueue and calls triage '''
    while self.check():
      try:
        for p in self.inQueue.get(block=True, timeout=1):
          self.triage(p)
          self.inQueue.task_done()
      except Queue.Empty,e:
        log.debug('Empty queue')
        pass
    self.finish()
    pass

  def triage(self, obj):
    ''' pile packets in the right state machine and call the processing 
      @param obj: the packet
    '''
    # check queues only
    if obj is None:
      self.stack.inbound.checkForExpectedPackets()
      self.stack.outbound.checkForExpectedPackets()
      return None
    
    # triage
    packet=obj[self.protocolName]
    pLen=len(packet.payload)
    if pLen == 0: # ignore acks and stuff 
      return None
    # real triage
    if self._isInbound(packet):
      log.debug('Packet is Inbound')
      if self.stack.inbound.checkState( packet ) and pLen > 0:
        log.debug('packet added')
    elif self._isOutbound(packet):
      log.debug('Packet is Outbound')
      if self.stack.outbound.checkState( packet ) and pLen > 0:
        log.debug('packet added')
    else:
      log.warning('This packet is nor outbound, nor inbound by my standards. Your network sniffer queuing system sucks.')
    return

  def finish(self):
    self.getInbound().finish()
    self.getOutbound().finish()
    log.info('Closing Stream - closing both sockets')

  def __str__(self):
    return "<TCPStream %s>"%str(self.connection)


class TCPStream(Stream):
  ''' Simple subclass for TCP packets '''
  def __init__(self, inQueue, connection ):
    Stream.__init__(self, inQueue, connection, protocolName='TCP')

  def _isInbound(self, packet):
    ''' check if the connection metadata corrects '''
    host,port = self.connection.local_address
    return host == packet.underlayer.dst and port == packet.dport

  def _isOutbound(self, packet):
    ''' check if the connection metadata corrects '''
    host,port = self.connection.local_address
    return host == packet.underlayer.src and port == packet.sport
    


    
    
    
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import argparse
import ctypes
import copy
import os
import logging
import pickle
import struct
import sys
import time
import threading
import Queue

# todo : replace by one empty shell of ours
from paramiko.transport import Transport

import ctypes_openssl
import ctypes_openssh
import output
import haystack 
import network
import utils

from engine import CIPHERS
#our impl
from paramiko_packet import Packetizer, PACKET_MAX_SIZE

log=logging.getLogger('sslsnoop-openssh')


class SessionStateFileWriter(output.FileWriter):
  def __init__(self,pid,folder='outputs'):
    FileWriter.__init__(self,'session_state',pid,folder)
  def writeToFile(self,instance):
    prefix = self.prefix
    filename = self.get_valid_filename()
    f = open(filename,"w")
    pickle.dump(instance,f)
    f.close()
    log.info ("[X] SSH session_state saved to file %s"%filename)
    return True


class Dummy:
  pass

class SessionCiphers():
  def __init__(self, session_state):
    self.session_state = session_state
    # two ciphers    
    self.receiveCtx = Dummy()
    self.sendCtx = Dummy()
    # read cipher MODE_IN == 0
    MODE=0
    for ctx in [self.receiveCtx, self.sendCtx ]:
      ctx.enc =  self.session_state.newkeys[MODE].enc
      ctx.mac =  self.session_state.newkeys[MODE].mac
      ctx.comp = self.session_state.newkeys[MODE].comp
      if MODE == 0:
        ctx.sshCtx = session_state.receive_context
      else:
        ctx.sshCtx = session_state.send_context
      ctx.sshCipher = ctx.sshCtx.cipher
      ctx.evpCtx    = ctx.sshCtx.evp
      ctx.evpCipher = ctx.sshCtx.evp.cipher
      # useful stuff
      ctx.name = ctx.sshCipher.name
      ctx.app_data  = ctx.evpCtx.app_data  #TODO delete
      ctx.cipher_data  = ctx.evpCtx.cipher_data #TODO delete
      # original key and IV are ctx.getKey() and ctx.getIV()
      # stateful AES_key key is at ctx.app_data.aes_ctx #&c->aes_ctx
      # stateful ctr counter is at ctx.app_data.aes_ctr
      log.debug("ctx.evpCipher type : %s"%(type(ctx.evpCipher)) )
      ctx.key_len  = ctx.evpCipher.key_len
      ctx.block_size  = ctx.evpCipher.block_size
      MODE+=1
    # 
    return

  def getCiphers(self):
    return self.receiveCtx, self.sendCtx
  
  def __str__(self):
    return "<SessionCiphers RECEIVE: '%s/%s' SEND: '%s/%s' >"%(self.receiveCtx.name,self.receiveCtx.mac.name,
                                                  self.sendCtx.name,self.sendCtx.mac.name ) 

class OpenSSHKeysFinder():
  ''' wrapper around a fork/exec to haystack StructFinder '''

  def __init__(self, pid, fullScan=False):
    self.pid = pid
    self.fullScan = fullScan
    return
  
  def findActiveSession(self, maxNum=1):
    ''' search for session_state '''
    outs = haystack.findStruct(self.pid, ctypes_openssh.session_state, debug=False)
    if outs is None:
      log.error("The session_state has not been found. maybe it's not OpenSSH ?")
      return None,None
    elif len(outs) > 1:
      log.warning("Mmmh, we found multiple session_state(%d). That is odd. I'll try with the first one."%(len(outs)))
    #
    session_state,addr = outs[0]
    return session_state, addr

  def refreshActiveSession(self, offset):
    ''' refresh session_state from a known address '''
    instance,validated = haystack.refreshStruct(self.pid, ctypes_openssh.session_state, offset)
    if not validated:
      log.error("The session_state has not been re-validated. You should look for it again.")
      return None,None
    return instance,offset
    
  def findActiveKeys(self, maxNum=1, offset=None):
    ''' wrap search or refresh and returns a session_state with pretty clothes'''
    if offset is None:
      session_state,addr = self.findActiveSession(maxNum)
    else:
      session_state,addr = self.refreshActiveSession(offset)
    if session_state is None:
      return None,None # raise Exception ... 
    log.debug('received session_state %s'%(session_state))
    ciphers = SessionCiphers(session_state)
    log.info('Active state ciphers : %s at 0x%lx'%(ciphers,addr))
    return ciphers,addr

  def save(self, instance):
    ssfw = SessionStateFileWriter(self.process.pid)
    ssfw.writeToFile(instance)
    return



class OpenSSHLiveDecryptatator(OpenSSHKeysFinder):
  ''' 
    Decrypt SSH traffic in live.
    This class only works on Live PID.
  '''
  def __init__(self, pid, sessionStateAddr=None, scapyThread = None, autoalign=True):
    OpenSSHKeysFinder. __init__(self, pid)
    self.scapy = scapyThread
    self.session_state_addr = sessionStateAddr
    self.connection = utils.getConnectionForPID(self.pid)
    self.inbound = Dummy()
    self.outbound = Dummy()
    self.autoalign = autoalign
    return
  
  def _initSniffer(self):
    ''' use existing sniffer or create a new one '''
    if self.scapy is None:
      self.scapy = utils.launchScapy()
    elif not self.scapy.thread.isAlive():
      self.scapy.thread.start()
    return
  
  def _initCiphers(self):
    ''' ptrace the ssh process to get sessions keys '''
    if self.session_state_addr is None:
      self.ciphers,self.session_state_addr = self.findActiveKeys()
    else:
      self.ciphers,self.session_state_addr = self.findActiveKeys(offset=self.session_state_addr)
    if self.ciphers is None:
      raise ValueError('Struct not found')
    log.debug('Ciphers loaded : %s %s'%(self.ciphers.receiveCtx.name,self.ciphers.sendCtx.name))
    return  
  
  def _initStream(self):
    ''' create a stream from scapy thread '''
    self.stream = self.scapy.makeStream(self.connection)
    self.inbound.state = self.stream.getInbound()
    self.outbound.state = self.stream.getOutbound()
    log.debug('Streams loaded')
    return
    
  def _initSSH(self):
    ''' plug sockets, packetizer and outputs together '''
    receiveCtx,sendCtx = self.ciphers.getCiphers()
    # Inbound
    log.debug('activate INBOUND packetizer')
    self.inbound.context = receiveCtx
    self.inbound.packetizer = Packetizer( self.inbound.state.getSocket() )
    self.inbound.packetizer.set_log(logging.getLogger('inbound.packetizer'))
    self.inbound.engine = self._attachEngine(self.inbound.packetizer, self.inbound.context )
    # Outbound
    log.debug('activate OUTBOUND packetizer')
    self.outbound.context = sendCtx
    self.outbound.packetizer = Packetizer(self.outbound.state.getSocket() )
    self.outbound.packetizer.set_log(logging.getLogger('outbound.packetizer'))
    self.outbound.engine = self._attachEngine(self.outbound.packetizer, self.outbound.context )
    return 

  def _attachEngine(self, packetizer, context):
    ''' activate the packetizer with a cipher engine '''
    # find Engine from engine.ciphers
    engine = CIPHERS[context.name](context) 
    log.debug( 'cipher:%s block_size: %d key_len: %d '%(context.name, context.block_size, context.key_len ) )
    mac = context.mac
    if mac is not None:
      mac_key    = mac.key 
      mac_engine = Transport._mac_info[mac.name]['class']
      mac_len = mac.mac_len
    # TODO : again , we need a stateful HMAC engine. 
    # we have disabled HMAC checking to get around.
    packetizer.set_inbound_cipher(engine, context.block_size, mac_engine, mac_len , mac_key)
    if context.comp.enabled != 0:
      name = context.comp.name
      compress_in = Transport._compression_info[name][1]
      log.debug('Switching on inbound compression ...')
      packetizer.set_inbound_compressor(compress_in())
    return engine
    
  def _initOutputs(self):
    ''' init output engine. File Writers.'''
    name = 'ssh-%s'%( utils.connectionToString(self.stream.connection, reverse=True) )
    self.inbound.filewriter = output.SSHStreamToFile(self.inbound.packetizer, self.inbound, name)

    name = 'ssh-%s'%( utils.connectionToString(self.stream.connection) )
    self.outbound.filewriter =  output.SSHStreamToFile(self.outbound.packetizer, self.outbound, name)
    log.debug('Outputs created')
    return 

  def _initWorker(self):
    ''' worker to poll on data for decryption '''
    self.worker = output.Supervisor()
    self.worker.add( self.inbound.state.getSocket() ,  self.inbound.filewriter.process  )
    self.worker.add( self.outbound.state.getSocket(), self.outbound.filewriter.process )
    log.debug('Worker created')
    return
        
  def _launchStreamProcessing(self):
    ''' run streams '''
    self.stream_t = threading.Thread(target=self.stream.run,name='stream' )
    self.stream_t.start()
    log.debug('Stream thread started.')
    return
    
  def run(self):
    ''' launch sniffer and decrypter threads '''
    # capure network before keys
    self._initSniffer()
    self._initStream()
    self._launchStreamProcessing()
    # we can get keys now
    self._initCiphers()
    self._initSSH()
    self._initOutputs()
    self._initWorker()
    log.info('Ready to catch some ssh traffic - please try `ls -l` in ssh if your just playing around...')
    if self.autoalign:
      log.info('trying to auto-align session keys and data')
      #log.info(self.ciphers.session_state.input.toString())
      threading.Thread(target=alignEncryption, args=(self.inbound, self.ciphers.session_state.incoming_packet), name='align inbound').start()
      threading.Thread(target=alignEncryption, args=(self.outbound, self.ciphers.session_state.outgoing_packet), name='align outbound').start()
    else:
      self.inbound.state.setActiveMode()
      self.outbound.state.setActiveMode()
    self.loop()
    log.info("done %s"%(self))
    return
    
  def loop(self):
    self.worker.run()
    return

  def refresh(self):
    log.warning('Refreshing Engine states')
    self._initCiphers()
    # refresh both engine
    receiveCtx,sendCtx = self.ciphers.getCiphers()
    # Inbound
    log.info('activate INBOUND receive')
    self.inbound.context = receiveCtx
    self.outbound.context = sendCtx
    self.inbound.engine.sync(self.inbound.context)
    self.outbound.engine.sync(self.outbound.context)
    pass


  def __str__(self):
    return "Decryption for pid %d, struct at 0x%lx"%(self.pid, self.session_state_addr)

def alignEncryption(way, packet_state, block=True):
  ''' 
    try to align the engine on the stream before activating it. 
    
    Question is: how do we know, at what offset of a packet the session keys state are.
    If there is no traffic beween the sniffer's start and alignTest(), the offset is 0 and obvious.
    but we need a full packet from headers to be able to decrypt it...
    
    I don't have a clue on how to align cipher state and data offset. Except 'brute force' _all_ offset.
    
    Best guess : try each packet and test first 4 bytes for a correct packet_size.
    That doesn't solve the problem for any 'active' ssh stream.
    Only solves offline pcap processing, if openssh was processing a packet.
    
    how do we know it's the start of a SSH Message ?
    well, msg[0:4] is the packet_size.  and a SSH Packet < PACKET_MAX_SIZE (0x00040000)
    that already gives us first byte 0x00 (1/255) * second byte (4/255) = 4/65335 . good odds.
    Still, not enough.
    the packet size must have :  (packet_size - (blocksize-4)) % blocksize ) = 0.
    Odds of that ? 1 / blocksize. 
    so odds are 1/262140 . Nice.  0.00038 %. 1/0x3fffc. 
    So basically, it's possible the alignement is made wrong. 
    But knowing that we _infer_ that the cipher state is 'before message {en,de}cryption'
    and that PACKET_MAX_SIZE == 0x40000, odds of 1/0x3fffc, for a message of maximum len 0x40000 
    means, there is no way in hell i got that wrong. 
      IF the cipher state has been captured between two messages.

    => reality kills theory - occurence of bad offset with valid tests has been seen in tests 
    so much for the stats. On big packets (1500) , 10 occurrences on 1100 packets . 
  '''
  # way.engine way.state
  name = threading.currentThread().name
  #import time # debug
  #time.sleep(5)
  log.debug('%s: trying to align on data'%(name))
  if packet_state.offset != packet_state.end:
    log.error("%s: openssh was in the middle of processing a packet. I can't deal with that "%(name))
    log.warning(packet_state.toString())
    way.state.setActiveMode() # it's gonna fail...
    return
  # head 
  try:
    data, qsize = way.state.getFirstPacketData(block=block)
    nbp = 1
    #if qsize > 1: # crowded traffic, lets cut to the point.
    #  for i in xrange(1, qsize/3):
    #    way.state.getFirstPacketData(block=block)
    #    nbp += 1
    #  log.info('dropping %d packets'%(qsize/2))
    #  data, qsize = way.state.getFirstPacketData(block=block)
    #  nbp += 1
  except Queue.Empty,e:
    way.state.setActiveMode() # it's on...
    return
  # rest
  while True:
    log.debug('%s: packet next %d'%(name, nbp))
    blocksize = way.engine.block_size
    # try to find a packetlen
    #for i in range(0, len(data)-blocksize, len(data) ): # check only start of packet
    for i in range(0, len(data)-blocksize, 1 ): # check all offsets
    # tests shows Message are often data[blocksize:] 
    # is that because data = data[len(data)-blocksize:] ???
    #for i in range(0, len(data)-blocksize, blocksize ): 
      #log.debug('trying index %d'%(i))
      header = way.engine.decrypt( data[i:i+blocksize] )
      packet_size = struct.unpack('>I', header[:4])[0]
      # reset engine to initial state
      way.engine.sync(way.context)
      if  0 < packet_size <= PACKET_MAX_SIZE:
        log.debug('%s: Auto align done: We found a acceptable packet size(%d) at %d on packet num %d'%(name, packet_size, i, nbp))
        if (packet_size - (blocksize-4)) % blocksize == 0 :
          log.info('%s: Auto align found : packet size(%d) at %d on packet num %d'%(name, packet_size, i, nbp))
          way.state.setActiveMode(data[i:])
          return
          # continue for test debug
        else:
          log.debug('%s: bad blocking packetsize %d is not correct for blocksize'%(name, (packet_size - (blocksize-4)) % blocksize))
    data = data[len(data)-blocksize:]
    #data = ''
    log.debug('%s: trying next packet '%(name))
    try:
      d, qsize = way.state.getFirstPacketData(block=False)
      data += d
      nbp += 1
    except Queue.Empty,e:
      log.warning('%s: no packets waiting for us after %d tries, offset is long gone... alignEncryption failed'%(name, nbp))
      way.state.setActiveMode() # it's gonna fail...
      return
    pass

class OpenSSHPcapDecrypt(OpenSSHLiveDecryptatator):
  ''' 
  Decrypt ssh traffic from a dumped session_state and a pcap capture.
  '''
  def __init__(self, pcapfilename, connection, ssfile):
    self.scapy = None
    self.session_state_addr = None
    self.inbound = Dummy()
    self.outbound = Dummy()
    self.autoalign = True
    # now...
    self.ssfile = ssfile
    self.pcapfilename = pcapfilename
    self.connection = connection

  def _initSniffer(self):
    ''' use existing sniffer or create a new one '''
    self.scapy = network.PcapFileSniffer(self.pcapfilename)
    sniffer = threading.Thread(target=self.scapy.run, name='scapy')
    self.scapy.thread = sniffer
    # do not launch the scapy thread before having made the stream, 
    # otherwise we are gonna loose packets ...
    log.debug('sniffer created: %s'%(self.scapy))
    return
  
  def _initCiphers(self):
    ''' ptrace the ssh process to get sessions keys '''
    inst = pickle.load(self.ssfile)
    self.session_state,self.session_state_addr = inst[0]
    self.ciphers = SessionCiphers(self.session_state)
    if self.ciphers is None:
      raise ValueError('Struct not found')
    log.debug('Ciphers loaded : %s %s'%(self.ciphers.receiveCtx.name,self.ciphers.sendCtx.name))
    return  
  
  def _launchStreamProcessing(self):
    OpenSSHLiveDecryptatator._launchStreamProcessing(self)
    # we can start scapy now, stream are in place
    self.scapy.thread.start()
    log.debug('sniffer thread started')
    return

  def __str__(self):
    return "Decryption of %s, struct at 0x%lx"%(self.pcapfilename, self.session_state_addr)


def launchLiveDecryption(pid, sniffer, addr=None): 
  ''' launch a live decryption '''
  # sniffer is a running thread
  # when ready, will have to launch tcpstream as a Thread
  decryptatator = OpenSSHLiveDecryptatator(pid, sessionStateAddr=addr, scapyThread=sniffer )
  decryptatator.run()
  return

def launchPcapDecryption(pcap, connection, ssfile): 
  ''' launch a decryption from a pcap file and a session state '''
  decryptatator = OpenSSHPcapDecrypt(pcap, connection, ssfile)
  decryptatator.run()
  return



# ============== move to scripts/sslsnoop-openssh


def argparser():
  parser = argparse.ArgumentParser(prog='sshsnoop', description='Live decription of Openssh traffic.')

  subparsers = parser.add_subparsers(help='sub-command help')
  live_parser = subparsers.add_parser('live', help='Decrypts traffic from a live PID.')
  live_parser.add_argument('pid', type=int, help='Target PID')
  live_parser.add_argument('--addr', type=str, help='active_context memory address')
  live_parser.add_argument('--debug', action='store_const', const=True, default=False, help='debug mode')
  live_parser.set_defaults(func=search)

  offline_parser = subparsers.add_parser('offline', help='Decrypts traffic from a pcap file, given a pickled session state.')
  offline_parser.add_argument('sessionstatefile', type=argparse.FileType('r'), help='File containing a pickled sessionstate.')
  offline_parser.add_argument('pcapfile', type=argparse.FileType('r'), help='Pcap file containing ssh traffic.')
  offline_parser.add_argument('src', type=str, help='SSH local host ip.')
  offline_parser.add_argument('sport', type=int, help='SSH source port.')
  offline_parser.add_argument('dst', type=str, help='SSH remote host ip.')
  offline_parser.add_argument('dport', type=int, help='SSH destination port.')
  offline_parser.add_argument('--debug', action='store_const', const=True, default=False, help='debug mode')
  offline_parser.set_defaults(func=searchOffline)
  return parser

def search(args):

  # we must have big privileges...
  if os.getuid() + os.geteuid() != 0:
    log.error("You must be root/using sudo to read memory and sniff traffic.")
    return
  
  pid = int(args.pid)
  log.debug("Target has pid %d"%pid)
  addr = None
  if args.addr != None:
    addr = int(args.addr,16)
  launchLiveDecryption(pid, None, addr=addr)
  sys.exit(0)
  return

def searchOffline(args):
  import utils 
  connection = utils.Connection(args.src,args.sport, args.dst,args.dport)
  launchPcapDecryption(args.pcapfile.name, connection, args.sessionstatefile)
  sys.exit(0)
  return


def main(argv):
  #logging.getLogger('network').setLevel(level=logging.INFO)
  #logging.getLogger('stream').setLevel(level=logging.INFO)
  #logging.getLogger('sslsnoop-openssh').setLevel(level=logging.DEBUG)

  if '--debug' in argv:
    logging.basicConfig(level=logging.DEBUG)    
    log.debug("==================- DEBUG MODE -================== ")  
  else:
    logging.basicConfig(level=logging.INFO)

  parser = argparser()
  opts = parser.parse_args(argv)
  try:
    opts.func(opts)
  except ImportError,e:
    log.error('Struct type does not exists.')
    print e

  return 0
  
if __name__ == "__main__":
  main(sys.argv[1:])




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import logging
import pickle
import sys
import socket
import threading

import argparse
import psutil

import output
import network
import openssh #OpenSSHLiveDecryptatator
from paramiko_packet import Packetizer

log = logging.getLogger('utils')



def connectionToString(connection, reverse=False):
  log.debug('make a string for %s'%(repr(connection)))
  if reverse:
    return "%s:%d-%s:%d"%(connection.remote_address[0],connection.remote_address[1], connection.local_address[0],connection.local_address[1]) 
  else:
    return "%s:%d-%s:%d"%(connection.local_address[0],connection.local_address[1],connection.remote_address[0],connection.remote_address[1]) 

def getConnectionForPID(pid):
  proc = psutil.Process(pid)
  return checkConnections(proc)

def checkConnections(proc):
  conns=proc.get_connections()
  conns = [ c for c in conns if c.status == 'ESTABLISHED']
  if len(conns) == 0 :
    return False
  elif len(conns) > 1 :
    log.warning(' %s has more than 1 connection ?'%(proc.name))
    return False
  elif conns[0].status != 'ESTABLISHED' :
    log.warning(' %s has no ESTABLISHED connections (1 %s)'%(conn[0].status))
    return False
  log.info('Found connection %s for %s'%(conns[0], proc.name))
  return conns[0]

class Connection:
  ''' Mimic psutils connection class '''
  def __init__(self, src, sport, dst, dport):
    self.local_address = (src,sport)
    self.remote_address = (dst, dport)
    self.status = 'ESTABLISHED'
  def __str__(self):
    return "%s:%d-%s:%d"%(self.local_address[0],self.local_address[1],self.remote_address[0],self.remote_address[1]) 



def launchScapy():
  from threading import Thread
  sshfilter = "tcp "
  soscapy = network.Sniffer(sshfilter)
  sniffer = Thread(target=soscapy.run)
  soscapy.thread = sniffer
  sniffer.start()
  return soscapy





def dumpPcapToFiles(pcapfile, connection, fname='raw'):
  # create the pcap reader
  soscapy = network.PcapFileSniffer(pcapfile)
  sniffer = threading.Thread(target=soscapy.run, name='scapy')
  soscapy.thread = sniffer
  # create output engines
  stream = soscapy.makeStream( connection ) 
  # prepare them
  stream.getInbound().setActiveMode()
  stream.getOutbound().setActiveMode()
  in_s = stream.getInbound().getSocket()
  out_s = stream.getOutbound().getSocket()
  win = output.RawPacketsToFile(in_s, fname+'-in.raw')
  wout = output.RawPacketsToFile(out_s, fname+'-out.raw')
  # make thread for all that
  st = threading.Thread(target=stream.run, name='stream')
  st.start()
  win_th = threading.Thread(target=win.run, name='inbound data writer')
  wout_th = threading.Thread(target=wout.run, name='outbound data writer')
  win_th.start()
  wout_th.start()
  # launch scapy and start the dumping
  sniffer.start()
  # wait for them
  for t in [sniffer, st, win_th, wout_th]:
    log.debug('waiting for %s'%(t.name))
    t.join()
    in_s.close()  
  return (fname+'-in.raw', fname+'-out.raw')


def rawFilesFromPcap(pcapfile, connection, fname='raw'):
  fin,fout = dumpPcapToFiles(pcapfile, connection, fname='raw')
  return file(fin), file(fout)


def writeToSocket(socket_,file_):
  while True:
    data = file_.read(4096)
    if len(data) == 0:
      break
    socket_.send(data)



def dump(args):
  connection = Connection(args.src,args.sport, args.dst,args.dport)
  fname = 'raw-%s'%(connection)
  args.pcapfile.close()
  dumpPcapToFiles(args.pcapfile.name, connection, fname)


def offline(args):
  connection = Connection(args.src,args.sport, args.dst,args.dport)
  fname = 'raw-%s'%(connection)
  args.pcapfile.close()
  decrypt = openssh.OpenSSHPcapDecrypt(args.pcapfile.name, connection, args.sessionstatefile)
  decrypt.run()
  log.info('Decrypt done -- ')







#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import argparse, os, logging, sys, time, pickle, struct, threading

import ctypes
import ctypes_openssh, ctypes_openssl
import output

from engine import CIPHERS
import haystack 


from output import FileWriter

log=logging.getLogger('openvpn')





class OpenvpnKeysFinder():
  ''' wrapper around a fork/exec to abouchet StructFinder '''

  def __init__(self, pid, fullScan=False):
    self.pid = pid
    self.fullScan = fullScan
    return
  
  def findCipherCtx(self, maxNum=3):
    ''' '''
    outs=haystack.findStruct(self.pid, ctypes_openssl.EVP_CIPHER_CTX, maxNum=maxNum)
    if outs is None:
      log.error("The session_state has not been found. maybe it's not OpenSSH ?")
      return None,None
    #
    return outs

  def refreshCipherCtx(self, offset):
    ''' '''
    instance,validated=haystack.refreshStruct(self.pid, ctypes_openssl.EVP_CIPHER_CTX, offset)
    if not validated:
      log.error("The session_state has not been re-validated. You should look for it again.")
      return None,None
    return instance,offset
    
  def findKeys(self, maxNum=3):
    ''' '''
    contexts = self.findCipherCtx(maxNum)
    log.debug('received %d EVP context '%( len(contexts)))
    self.printCiphers(contexts)
    return contexts

  def printCiphers(self, cipherContexts):
    for ctx, addr in cipherContexts:
      print 'Cipher : %s-%d'%(ctypes_openssl.getCipherName(ctx.cipher.nid), 8*ctx.cipher.key_len)
    return 
    
  def save(self, instance):
    # pickle ?
    ssfw=FileWriter(self.process.pid)
    ssfw.writeToFile(instance)
    return



def connectionToString(connection, reverse=False):
  log.debug('make a string for %s'%(repr(connection)))
  if reverse:
    return "%s:%d-%s:%d"%(connection.remote_address[0],connection.remote_address[1], connection.local_address[0],connection.local_address[1]) 
  else:
    return "%s:%d-%s:%d"%(connection.local_address[0],connection.local_address[1],connection.remote_address[0],connection.remote_address[1]) 



class OpenvpnLiveDecryptatator(OpenvpnKeysFinder):
  ''' Decrypt SSH traffic in live '''
  def __init__(self, pid, sessionStateAddr=None, stream = None, scapyThread = None, serverMode=None, fullScan=False, maxNum=1):
    OpenSSHKeysFinder. __init__(self, pid, fullScan=fullScan)
    self.stream = stream
    self.scapy = scapyThread
    #self.serverMode=serverMode
    self.maxNum = maxNum
    self.session_state_addr=sessionStateAddr
    self.inbound=dict()
    self.outbound=dict()
    return
  
  def run(self):
    ''' launch sniffer and decrypter threads '''
    if self.scapy is None:
      from finder import launchScapy, getConnectionForPID
      self.scapy=launchScapy()
      conn = getConnectionForPID(self.pid)
      self.stream = self.scapy.makeStream(conn)
    elif not self.scapy.thread.isAlive():
      self.scapy.thread.start()
    # ptrace ssh
    if self.session_state_addr is None:
      self.ciphers,self.session_state_addr=self.findActiveKeys(maxNum = self.maxNum)
    else:
      self.ciphers,self.session_state_addr=self.findActiveKeys(offset=self.session_state_addr)
    if self.ciphers is None:
      raise ValueError('Struct not found')
    # unstop() the process
    ### Forked no useful self.process.cont()
    # process is running... sniffer is listening
    log.info('Please make some ssh  traffic')  
    self.init()
    self.loop()
    log.info("done for pid %d, struct at 0x%lx"%(self.process.pid, self.session_state_addr))
    return
    
  def init(self):
    ''' plug sockets, packetizer and outputs together '''
    receiveCtx,sendCtx = self.ciphers.getCiphers()
    # Inbound
    log.info('activate INBOUND receive')
    self.inbound['context'] = receiveCtx
    self.inbound['socket'] = self.stream.getInbound().getSocket()
    self.inbound['packetizer'] = Packetizer( self.inbound['socket'] )
    self.inbound['packetizer'].set_log(logging.getLogger('inbound.packetizer'))
    self.inbound['engine'] = self.activate_cipher(self.inbound['packetizer'], receiveCtx )
    name = 'ssh-%s'%( connectionToString(self.stream.connection) )
    self.inbound['filewriter'] =  output.SSHStreamToFile(self.inbound['packetizer'], self.inbound, name)

    # out bound
    log.info('activate OUTBOUND send')
    self.outbound['context'] = sendCtx
    self.outbound['socket'] = self.stream.getOutbound().getSocket()
    self.outbound['packetizer'] = Packetizer(self.outbound['socket'])
    self.outbound['packetizer'].set_log(logging.getLogger('outbound.packetizer'))
    self.outbound['engine'] = self.activate_cipher(self.outbound['packetizer'], self.outbound['context'] )
    name = 'ssh-%s'%( connectionToString(self.stream.connection, reverse=True) )
    self.outbound['filewriter'] =  output.SSHStreamToFile(self.outbound['packetizer'], self.outbound, name)
    
    # worker to watch out for data for decryption
    self.worker = output.Supervisor()
    self.worker.add( self.inbound['socket'],  self.inbound['filewriter'].process  )
    self.worker.add( self.outbound['socket'], self.outbound['filewriter'].process )
    ## run streams
    self.stream.getInbound().setActiveMode()
    self.stream.getOutbound().setActiveMode()
    threading.Thread(target=self.stream.run).start()
    return
    
  def loop(self):
    self.worker.run()
    return

  def activate_cipher(self, packetizer, context):
    "switch on newly negotiated encryption parameters for inbound traffic"
    #packetizer.set_log(log)
    #packetizer.set_hexdump(True)
    # find Engine from engine.ciphers
    engine = CIPHERS[context.name](context) 
    log.debug( 'cipher:%s block_size: %d key_len: %d '%(context.name, context.block_size, context.key_len ) )
    #print engine, type(engine)
    mac = context.mac
    if mac is not None:
      mac_key    = mac.key #XXX TODO
      mac_engine = Transport._mac_info[mac.name]['class']
      mac_len = mac.mac_len
    # again , we need a stateful HMAC engine. 
    # we disable HMAC checking to get around.
    # fix our engines in packetizer
    ## packetizer.set_hexdump(True)
    packetizer.set_inbound_cipher(engine, context.block_size, mac_engine, mac_len , mac_key)
    if context.comp.enabled != 0:
      name = context.comp.name
      compress_in = Transport._compression_info[name][1]
      log.debug('Switching on inbound compression ...')
      packetizer.set_inbound_compressor(compress_in())
    #ok
    return engine
  
  def refresh(self):
    log.warning('Refreshing Engine states')
    # ptrace ssh
    if self.session_state_addr is None:
      self.ciphers,self.session_state_addr=self.findActiveKeys(maxNum = self.maxNum)
    else:
      self.ciphers,self.session_state_addr=self.findActiveKeys(offset=self.session_state_addr)
    if self.ciphers is None:
      raise ValueError('Struct not found')
    # refresh both engine
    receiveCtx,sendCtx = self.ciphers.getCiphers()
    # Inbound
    log.info('activate INBOUND receive')
    self.inbound['context'] = receiveCtx
    self.outbound['context'] = sendCtx
    self.inbound['engine'].sync(self.inbound['context'])
    self.outbound['engine'].sync(self.outbound['context'])
    pass



def argparser():
  parser = argparse.ArgumentParser(prog='sshsnoop', description='Live decription of Openssh traffic.')
  parser.add_argument('pid', type=int, help='Target PID')
  parser.add_argument('--addr', type=str, help='active_context memory address')
  parser.add_argument('--debug', action='store_const', const=True, default=False, help='debug mode')
  parser.set_defaults(func=search)
  return parser

def search(args):
  if args.debug:
    logging.basicConfig(level=logging.DEBUG)    
    log.debug("==================- DEBUG MODE -================== ")
  pid=int(args.pid)
  log.info("Target has pid %d"%pid)
  if args.addr != None:
    offset = int(args.addr,16)
  
  test = OpenvpnKeysFinder(args.pid, fullScan=True)
  test.findKeys()
  
  #decryptatator=OpenSSHLiveDecryptatator(pid, sessionStateAddr=sessionStateAddr, serverMode=serverMode)
  #decryptatator.run()
  #log.info("done for pid %d, struct at 0x%lx"%(pid,decryptatator.session_state_addr))
  sys.exit(0)
  return -1


def main(argv):
  logging.basicConfig(level=logging.INFO)

  # we must have big privileges...
  if os.getuid() + os.geteuid() != 0:
    log.error("You must be root/using sudo to read memory and sniff traffic.")
    return

  parser = argparser()
  opts = parser.parse_args(argv)
  try:
    opts.func(opts)
  except ImportError,e:
    log.error('Struct type does not exists.')
    print e

  return 0
  



if __name__ == "__main__":
  main(sys.argv[1:])





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import logging,os,socket,select, sys,time
import multiprocessing, Queue
import scapy.config

from lrucache import LRUCache

import stream

log = logging.getLogger('network')

QUEUE_SIZE = 15000

def hexify(data):
  s=''
  for i in range(0,len(data)):
    s+="%02x"% ord(data[i])
    if i%16==15:
      s+="\r\n"
    elif i%2==1:
      s+=" "
  #s+="\r\n"
  return s

class Sniffer():
  worker=None
  def __init__(self,filterRules='tcp', packetCount=0, timeout=None):
    ''' 
    This sniffer can run in a thread. But it should be one of the few thread running (BPL) 
    
    @param filterRules: a pcap compatible filter string 
    @param packetCount: 0/Unlimited or packet capture limit
    @param timeout: None/Unlimited or stop after
    '''
    # set scapy to use native pcap instead of SOCK_RAW
    scapy.config.conf.use_pcap = True
    ## if using SOCK_RAW, we need to mess with filter to up the capture size higher than 1514/1600 bytes
    #maxSize="\' -s \'0xffff" # abusing scapy-to-tcpdump string format 
    #self.filterRules=filterRules + maxSize
    self.filterRules = filterRules
    self.packetCount = packetCount
    self.timeout = timeout
    #
    self.streams = {}
    self._running_thread = None
    return
  

  def run(self):
    # scapy - with config initialised
    #scapy.sendrecv.sniff(count=self.packetCount,timeout=self.timeout,store=0,filter=self.filterRules,prn=self.cbSSHPacket)
    from scapy.all import sniff
    log.info('Using L2listen = %s'%(scapy.config.conf.L2listen)) 
    # XXX TODO, define iface from saddr and daddr // scapy.all.read_routes()
    sniff(count=self.packetCount, timeout=self.timeout, store=0, filter=self.filterRules, prn=self.enqueue, iface='any')
    log.warning('============ SNIFF Terminated ====================')
    return

  def hasStream(self, packet):
    ''' checks if the stream has a queue '''
    (shost,sport,dhost,dport) = getConnectionTuple(packet)
    return (shost,sport,dhost,dport) in self.streams

  def getStream(self, packet):
    ''' returns the queue for that stream '''
    if self.hasStream(packet):
      return self.streams[getConnectionTuple(packet)]
    #log.debug('Packet with no such connection %s %s %s %s'%(getConnectionTuple(packet)))
    return None,None

  def addStream(self, connection):
    ''' forget that stream '''
    shost,sport = connection.local_address
    dhost,dport = connection.remote_address
    if shost.startswith('127.') or shost.startswith('::1'):
      log.warning('=============================================================')
      log.warning('scapy is gonna truncate big packet on the loopback interface.')
      log.warning('please change your test params,or use offline mode with pcap.')
      log.warning('=============================================================')
    #q = multiprocessing.Queue(QUEUE_SIZE)
    q = Queue.Queue(QUEUE_SIZE)
    st = stream.TCPStream(q, connection)
    #save in both directions
    self.streams[(shost,sport,dhost,dport)] = (st,q)
    self.streams[(dhost,dport,shost,sport)] = (st,q)
    return st
  
  def dropStream(self, packet):
    ''' forget that stream '''
    if self.hasStream(packet):
      (shost,sport,dhost,dport) = getConnectionTuple(packet)
      if (shost,sport,dhost,dport) in self.streams:
        del self.streams[(shost,sport,dhost,dport)]
      if (dhost,dport,shost,sport) in self.streams:
        del self.streams[(dhost,dport,shost,sport)]
      log.info('Dropped %s,%s,%s,%s from valid connections.'%(shost,sport,dhost,dport))
    return None
    
  def enqueue(self, packet):
    st,q = self.getStream(packet)
    if q is None:
      return
    try:
      log.debug('Queuing a packet from %s:%s\t-> %s:%s'%(getConnectionTuple(packet)))
      q.put_nowait(packet)
    except Queue.Full:
      log.warning('a Queue is Full (%d). lost packet for %s'%(q.qsize(), repr(st.connection)))
      self.dropStream( packet )
    except Exception,e:
      log.error(e)
    return

  def makeStream(self, connection):
    ''' create a TCP Stream recognized by sniffer 
      the Stream can be used to read captured data
      
      The Stream should be run is a thread or a subprocess (better because of GIL).
    '''
    shost,sport = connection.local_address
    dhost,dport = connection.remote_address
    if (shost,sport,dhost,dport) in self.streams:
      raise ValueError('Stream already exists')
    tcpstream = self.addStream(connection)
    log.debug('Created a TCPStream for %s'%(tcpstream))
    return tcpstream


def getConnectionTuple(packet):
  ''' Supposedly an IP/IPv6 model'''
  try:
    shost  = packet.payload.src
    sport  = packet.payload.payload.sport
    dhost  = packet.payload.dst
    dport  = packet.payload.payload.dport
  except Exception, e:
    log.debug("%s - %s"% (type(packet), packet.show2()))
    raise e
  return (shost,sport,dhost,dport)  
 


class PcapFileSniffer(Sniffer):
  ''' Use scapy's offline mode to simulate network by reading a pcap file.
  '''
  def __init__(self, pcapfile, filterRules='tcp', packetCount=0):
    Sniffer.__init__(self, filterRules=filterRules, packetCount=packetCount)
    self.pcapfile = pcapfile
  def run(self):
    from scapy.all import sniff
    sniff(store=0, prn=self.enqueue, offline=self.pcapfile)
    log.info('Finishing the pcap reading')
    for v, k in list(self.streams.items()):
      st,q = k
      if not q.empty():
        log.debug('waiting on %s'%(repr(v)))
        q.join()
      del self.streams[v]
      del q
      st.pleaseStop()

    log.info('============ SNIFF Terminated ====================')

    return


    
    
    













#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import os
import logging
import subprocess
import sys
import time

import psutil

import openssh
import openssl
import utils

log=logging.getLogger('finder')

class Args:
  pid=None
  memfile=None
  memdump=None
  debug=None

def parseSSL(pid, sniffer):
  args=Args()
  args.pid = pid
  return openssl.search(args)
  



_targets={
    'ssh': openssh.launchLiveDecryption,
    'ssh-agent': parseSSL,
    'sshd': parseSSL, # sshd a SSL keys too
    #'firefox': []
    }

Processes = []

def buildTuples(targets):
  rets=[]
  for proc in psutil.process_iter():
    if proc.name in targets:
      rets.append( (proc.pid,proc) )
  return rets

    
def pgrep(name):
  pids=[]
  for proc in psutil.process_iter():
    if proc.name == name:
      pids.append(name)
  return pids

def makeFilter(conn):
  # [connection(fd=3, family=2, type=1, local_address=('192.168.1.101', 36386), remote_address=('213.186.33.2', 22), status='ESTABLISHED')]
  pcap_filter = "host %s and port %s and host %s and port %s" %(conn.local_address[0],conn.local_address[1] ,
                            conn.remote_address[0],conn.remote_address[1]  )
  return pcap_filter


    
def runthread(callable, sniffer, proc,conn):
  ##from multiprocessing import Process
  ##p = Process(target=s1.run)
  from threading import Thread
  args = (proc.pid, sniffer)
  p = Thread(target=callable, args=args)
  p.start()
  Processes.append(p)
  log.info('Thread launched')
  return 
    

def main(argv):
  logging.basicConfig(level=logging.INFO)
  #logging.getLogger('model').setLevel(logging.INFO)


  # we must have big privileges...
  if os.getuid() + os.geteuid() != 0:
    log.error("You must be root/using sudo to read memory and sniff traffic. So there's no point in going further")
    return
  
  if not os.access('outputs', os.X_OK) :
    os.mkdir('outputs/')
  
  options=buildTuples(_targets)
  threads=[]
  forked=0
  # get sniffer up
  sniffer = utils.launchScapy()  
  for pid,proc in options:
    log.info("Searching in %s/%d memory"%(proc.name,proc.pid))
    conn = utils.checkConnections(proc)
    if not conn and 'ssh-agent' != proc.name:
      continue
    log.info('Adding this pid to watch list')
    runthread(_targets[proc.name], sniffer, proc,conn)
    
    forked+=1
    log.info('Subprocess launched on pid %d'%(proc.pid))

  for p in Processes:
    p.join()
  time.sleep(5)
  log.info(' ============== %d process forked. look into outputs/ for data '%(forked))
  sys.exit(0)
  return 0


if __name__ == "__main__":
  main(sys.argv[1:])


  
  
  



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import logging, os, subprocess

log=logging.getLogger('preprocess')

'''
OPENSSL_ARGS= ['-DOPENSSL_THREADS', '-D_REENTRANT', '-DDSO_DLFCN', '-DHAVE_DLFCN_H', 
    '-DL_ENDIAN', '-DTERMIO', '-O3', '-fomit-frame-pointer', '-Wall', '-DOPENSSL_BN_ASM_PART_WORDS',
    '-DOPENSSL_IA32_SSE2', '-DSHA1_ASM', '-DMD5_ASM', '-nostdinc++', '-DRMD160_ASM', '-DAES_ASM' ]
'''
OPENSSL_ARGS= [
  '-I.', 
  '-I/usr/include/', 
  '-I/usr/include/openssl', 
#  '-I/home/jal/Compil/sslsnoop/biblio//include', 
]

'''
NSS_ARGS= ['-Wall', '-pipe', '-O2', '-fno-strict-aliasing', '-g', '-ansi', '-D_POSIX_SOURCE', 
  '-D_BSD_SOURCE', '-D_XOPEN_SOURCE', '-fPIC', '-D_XOPEN_SOURCE', '-DLINUX1_2', '-Di386', '-DLINUX2_1', 
  '-Wno-switch', '-pipe', '-DLINUX', '-Dlinux', '-DHAVE_STRERROR', 
  '-DXP_UNIX', '-DSHLIB_SUFFIX="so"', '-DSHLIB_PREFIX="lib"', '-DSHLIB_VERSION="3"', '-DSOFTOKEN_SHLIB_VERSION="3"', 
  '-UDEBUG', '-DNDEBUG', '-D_REENTRANT', '-DNSS_ENABLE_ECC', '-DUSE_UTIL_DIRECTLY', 
'''
NSS_ARGS= [
  '-I.', 
  '-I/usr/include/', 
  '-I/usr/include/nspr', 
  '-I/home/jal/Compil/sslsnoop/biblio/nss-3.12.8/mozilla/dist/include', 
  '-I/home/jal/Compil/sslsnoop/biblio/nss-3.12.8/mozilla/dist/public/nss', 
  '-I/home/jal/Compil/sslsnoop/biblio/nss-3.12.8/mozilla/dist/private/nss',
  '-I/home/jal/Compil/sslsnoop/biblio/nss-3.12.8/mozilla/dist/public/dbm',
  '-I/home/jal/Compil/sslsnoop/biblio/nss-3.12.8/mozilla/security/nss/lib/ssl/'
  ]

class Preprocess:
  '''
    Make a Preprocessed File from a C file
    
    @param cfile: the desired input c file name
    @param preprocessed: the preprocessed file name
  '''
  #sslInc#export INCLUDES="-I$PWD/biblio/openssl-0.9.8o/crypto/ -I$PWD/biblio/openssl-0.9.8o/"

  def __init__(self, cfile, preprocessed, cppargs=None, gcc='gcc'):
    self.cfile = cfile
    self.preprocessed = preprocessed
    self.cppargs = cppargs
    self.gcc = 'gcc'
  
  def run(self):
    cmd_line = [self.gcc,  '-x', 'c++', self.cfile, '-E', '-P']
    cmd_line.extend(self.cppargs)
    cmd_line.extend(['-o', self.preprocessed])
    print ' '.join(cmd_line)
    p = subprocess.Popen(cmd_line, stdin=None, stdout=subprocess.PIPE, close_fds=True)
    p.wait()
    build = p.stdout.read().strip()
    if len(build) == 0:
      log.info("GENERATED %s - please correct source code gccxml if next step fails"%(self.preprocessed))
    else:
      log.info(build)
    return len(build)



def process(cfile, preprocessed):
  #p = Preprocess(cfile, preprocessed, OPENSSL_ARGS)
  p = Preprocess(cfile, preprocessed, NSS_ARGS)
  return p.run()


